@@ -5,7 +5,7 @@ from .Utilities import *
 # Public function(s)
 #####################################################
 
-def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[], delimiter="\t", comment_prefix="#", compression_type=None, num_threads=1, num_cols_per_chunk=None, num_rows_per_write=None, tmp_dir_path=None, verbose=False):
+def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[], delimiter="\t", comment_prefix="#", compression_type=None, num_parallel=1, num_cols_per_chunk=None, num_rows_per_write=None, tmp_dir_path=None, verbose=False):
     if type(delimiter) != str:
         raise Exception("The delimiter value must be a string.")
 
@@ -23,8 +23,6 @@ def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[],
         if comment_prefix == "":
             comment_prefix = None
 
-    print_message(f"Converting from {delimited_file_path}", verbose)
-
     # Guess an optimal value for this parameter, if not specified.
     if not num_rows_per_write:
         raw_num_rows = 0
@@ -32,8 +30,13 @@ def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[],
             for line in in_file:
                 raw_num_rows += 1
 
-        num_rows_per_write = ceil(raw_num_rows / (num_threads * 2) + 1)
+        num_rows_per_write = ceil(raw_num_rows / (num_parallel * 2) + 1)
 
+    if num_parallel > 1:
+        global joblib
+        joblib = __import__('joblib', globals(), locals())
+
+    print_message(f"Converting from {delimited_file_path}", verbose)
     tmp_dir_path_chunks, tmp_dir_path_outputs, tmp_dir_path_indexes = prepare_tmp_dirs(tmp_dir_path)
 
     # Get column names. Remove any leading or trailing white space around the column names.
@@ -54,15 +57,15 @@ def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[],
 
     # Iterate through the lines to summarize each column.
     print_message(f"Summarizing each column in {delimited_file_path}", verbose)
-    if num_threads == 1:
+    if num_parallel == 1:
         chunk_results = [parse_columns_chunk(delimited_file_path, delimiter, comment_prefix, 0, num_cols, num_rows_per_write, compression_type, verbose)]
     else:
         # Guess an optimal value for this parameter, if not specified.
         if not num_cols_per_chunk:
-            num_cols_per_chunk = ceil(num_cols / (num_threads * 2) + 1)
+            num_cols_per_chunk = ceil(num_cols / (num_parallel * 2) + 1)
 
         column_chunk_indices = generate_chunk_ranges(num_cols, num_cols_per_chunk)
-        chunk_results = Parallel(n_jobs=num_threads)(delayed(parse_columns_chunk)(delimited_file_path, delimiter, comment_prefix, column_chunk[0], column_chunk[1], num_rows_per_write, compression_type, verbose) for column_chunk in column_chunk_indices)
+        chunk_results = joblib.Parallel(n_jobs=num_parallel, mmap_mode=None)(joblib.delayed(parse_columns_chunk)(delimited_file_path, delimiter, comment_prefix, column_chunk[0], column_chunk[1], num_rows_per_write, compression_type, verbose) for column_chunk in column_chunk_indices)
 
     # Summarize the column sizes and types across the chunks.
     column_sizes = []
@@ -87,13 +90,13 @@ def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[],
         raise Exception(f"A header row but no data rows were detected in {delimited_file_path}")
 
     print_message(f"Parsing chunks of {delimited_file_path} and saving to temp directory ({tmp_dir_path_chunks})", verbose)
-    line_lengths_dict = get_line_lengths_dict(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, column_compression_dicts, num_rows, num_threads, num_rows_per_write, verbose)
+    line_lengths_dict = get_line_lengths_dict(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, column_compression_dicts, num_rows, num_parallel, num_rows_per_write, verbose)
 
     print_message(f"Saving meta files for {f4_file_path}", verbose)
     write_meta_files(tmp_dir_path_outputs, column_sizes, line_lengths_dict, column_names, column_types, compression_type, column_compression_dicts, num_rows)
 
     print_message(f"Combining all data into a single file for {delimited_file_path}", verbose)
-    combine_into_single_file(tmp_dir_path_chunks, tmp_dir_path_outputs, f4_file_path, num_threads, num_rows_per_write)
+    combine_into_single_file(tmp_dir_path_chunks, tmp_dir_path_outputs, f4_file_path, num_parallel, num_rows_per_write)
 
     if index_columns:
         build_indexes(f4_file_path, index_columns, tmp_dir_path_indexes, verbose)
@@ -104,6 +107,123 @@ def convert_delimited_file(delimited_file_path, f4_file_path, index_columns=[],
 
     print_message(f"Done converting {delimited_file_path} to {f4_file_path}", verbose)
 
+def transpose(f4_src_file_path, f4_dest_file_path, num_parallel=1, tmp_dir_path=None, verbose=False):
+    if num_parallel > 1:
+        global joblib
+        joblib = __import__('joblib', globals(), locals())
+
+    print_message(f"Transposing {f4_src_file_path} to {f4_dest_file_path}", verbose)
+
+    if tmp_dir_path:
+        makedirs(tmp_dir_path, exist_ok=True)
+    else:
+        tmp_dir_path = mkdtemp()
+
+    tmp_dir_path = fix_dir_path_ending(tmp_dir_path)
+    makedirs(tmp_dir_path, exist_ok=True)
+
+    print_message("got here0")
+    with initialize(f4_src_file_path) as src_file_data:
+        print_message("got here1")
+        column_names, column_type_dict, column_coords_dict, bigram_size_dict = get_column_meta(src_file_data, set(), [])
+        print_message("got here2")
+        tmp_tsv_file_path = tmp_dir_path + "tmp.tsv.gz"
+
+        if num_parallel == 1:
+            col_coords = [column_coords_dict[name] for name in column_names]
+            save_transposed_line_to_temp(src_file_data.data_file_path, tmp_tsv_file_path, column_names, col_coords, bigram_size_dict, tmp_dir_path, verbose)
+        else:
+            print_message("got here3")
+            col_index_chunks = list(split_integer_list_into_chunks(list(range(1, get_num_cols(f4_src_file_path))), num_parallel))
+            col_index_chunks.insert(0, [0])
+            col_name_chunks = []
+            col_coords_chunks = []
+
+            print_message("got here4")
+            for col_index_chunk in col_index_chunks:
+                col_name_chunks.append([column_names[i] for i in col_index_chunk])
+                col_coords_chunks.append([column_coords_dict[column_names[i]] for i in col_index_chunk])
+
+            print_message("got here5")
+            joblib.Parallel(n_jobs=num_parallel, mmap_mode=None)(
+                joblib.delayed(save_transposed_line_to_temp)(src_file_data.data_file_path, f"{tmp_dir_path}{chunk_number}", col_name_chunks[chunk_number], col_coords_chunks[chunk_number], bigram_size_dict, tmp_dir_path, verbose) for
+                chunk_number in range(len(col_index_chunks)))
+
+            with gzip.open(tmp_tsv_file_path, "wb", compresslevel=1) as tmp_tsv_file:
+                for i in range(0, num_parallel + 1):
+                    chunk_file_path = f"{tmp_dir_path}{i}"
+
+                    with gzip.open(chunk_file_path, "rb", compresslevel=1) as chunk_file:
+                        for line in chunk_file:
+                            tmp_tsv_file.write(line)
+
+                    remove(chunk_file_path)
+
+        print_message(f"Converting temp file at {tmp_tsv_file_path} to {f4_dest_file_path}", verbose)
+        convert_delimited_file(tmp_tsv_file_path, f4_dest_file_path, compression_type=src_file_data.decompression_type, num_parallel=num_parallel, verbose=verbose)
+        remove(tmp_tsv_file_path)
+
+def inner_join(f4_left_src_file_path, f4_right_src_file_path, join_column, f4_dest_file_path, num_parallel=1, tmp_dir_path=None, verbose=False):
+    join_column = join_column.encode()
+
+    print_message(f"Inner joining {f4_left_src_file_path} and {f4_right_src_file_path} based on the {join_column} column, saving to {f4_dest_file_path}", verbose)
+
+    if tmp_dir_path:
+        makedirs(tmp_dir_path, exist_ok=True)
+    else:
+        tmp_dir_path = mkdtemp()
+
+    tmp_dir_path = fix_dir_path_ending(tmp_dir_path)
+    makedirs(tmp_dir_path, exist_ok=True)
+    tmp_tsv_file_path = tmp_dir_path + "tmp.tsv.gz"
+
+    with initialize(f4_left_src_file_path) as left_file_data:
+        with initialize(f4_right_src_file_path) as right_file_data:
+            left_column_names, left_column_type_dict, left_column_coords_dict, left_bigram_size_dict = get_column_meta(left_file_data, set(), [])
+            right_column_names, right_column_type_dict, right_column_coords_dict, right_bigram_size_dict = get_column_meta(right_file_data, set(), [])
+            #TODO: Add error checking to make sure join_column is present in left and right.
+
+            left_values = parse_values_in_column(left_file_data, join_column, left_column_coords_dict[join_column], left_bigram_size_dict)
+            right_values = parse_values_in_column(right_file_data, join_column, right_column_coords_dict[join_column], right_bigram_size_dict)
+            common_values = set(left_values) & set(right_values)
+
+            right_columns_to_save = [name for name in right_column_names if name != join_column]
+
+            right_index_dict = {}
+            for i, value in enumerate(right_values):
+                if value in common_values:
+                    right_index_dict.setdefault(value, []).append(i)
+
+            # TODO: It would be better to save in zstandard format. But the Python package
+            #      doesn't support reading line by line. Come up with a better way than gzip,
+            #      which is very slow.
+            #TODO: Parallelize this?
+            with gzip.open(tmp_tsv_file_path, "wb", compresslevel=1) as tmp_tsv_file:
+                tmp_tsv_file.write(b"\t".join(left_column_names + right_columns_to_save) + b"\n")
+
+                left_parse_function = get_parse_row_values_function(left_file_data)
+                right_parse_function = get_parse_row_values_function(right_file_data)
+                left_column_coords = [left_column_coords_dict[name] for name in left_column_names]
+                right_column_coords = [right_column_coords_dict[name] for name in right_columns_to_save]
+
+                for left_row_index, left_value in enumerate(left_values):
+                    if left_value in common_values:
+                        for right_row_index in right_index_dict[left_value]:
+                            left_save_values = left_parse_function(left_file_data, left_row_index, left_column_coords, bigram_size_dict=left_bigram_size_dict)
+                            right_save_values = right_parse_function(right_file_data, right_row_index, right_column_coords, bigram_size_dict=right_bigram_size_dict)
+
+                            tmp_tsv_file.write(b"\t".join(left_save_values + right_save_values) + b"\n")
+
+            # TODO: Expand this logic for all compression types. Make sure to document it.
+            compression_type = None
+            if left_file_data.decompression_type == "zstd" or right_file_data.decompression_type == "zstd":
+                compression_type = "zstd"
+
+            print_message(f"Converting temp file at {tmp_tsv_file_path} to {f4_dest_file_path}", verbose)
+            convert_delimited_file(tmp_tsv_file_path, f4_dest_file_path, compression_type=compression_type, num_parallel=num_parallel, verbose=verbose)
+
+            remove(tmp_tsv_file_path)
+
 def build_indexes(f4_file_path, index_columns, tmp_dir_path, verbose=False):
     if isinstance(index_columns, str):
         build_one_column_index(f4_file_path, index_columns, tmp_dir_path, verbose)
@@ -189,6 +309,7 @@ def parse_columns_chunk(delimited_file_path, delimiter, comment_prefix, start_in
             line = line.rstrip(b"\n")
 
             line_items = line.split(delimiter)
+
             for i in range(start_index, end_index):
                 column_sizes_dict[i] = max([column_sizes_dict[i], len(line_items[i])])
 
@@ -260,22 +381,18 @@ def parse_columns_chunk(delimited_file_path, delimiter, comment_prefix, start_in
                     #column_compression_dicts[i]["map"][bigram] = int2ba(j, length = length).to01()
                     column_compression_dicts[i]["map"][bigram] = j.to_bytes(length = num_bytes, byteorder = "big")
 
-                #TODO: Does this work? If not, do we need to iterate the file again?
                 column_sizes_dict[i] = column_max_length_dict[i] * num_bytes
-                # for unique_value in unique_values:
-                #     compressed_length = len(compress_using_2_grams(unique_value, column_compression_dicts[i]["map"]))
-                #     column_sizes_dict[i] = max(column_sizes_dict[i], compressed_length)
 
     return column_sizes_dict, column_types_dict, column_compression_dicts, num_rows
 
-def get_line_lengths_dict(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, num_rows, num_threads, num_rows_per_write, verbose):
-    if num_threads == 1:
+def get_line_lengths_dict(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, num_rows, num_parallel, num_rows_per_write, verbose):
+    if num_parallel == 1:
         line_lengths_dict = write_rows_chunk(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, 0, 0, num_rows, num_rows_per_write, verbose)
     else:
-        row_chunk_indices = generate_chunk_ranges(num_rows, ceil(num_rows / num_threads) + 1)
+        row_chunk_indices = generate_chunk_ranges(num_rows, ceil(num_rows / num_parallel) + 1)
 
-        # Find the line length.
-        line_lengths_dicts = Parallel(n_jobs=num_threads)(delayed(write_rows_chunk)(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, i, row_chunk[0], row_chunk[1], num_rows_per_write, verbose) for i, row_chunk in enumerate(row_chunk_indices))
+        # We are doing the import here because it is slow.
+        line_lengths_dicts = joblib.Parallel(n_jobs=num_parallel, mmap_mode=None)(joblib.delayed(write_rows_chunk)(delimited_file_path, tmp_dir_path_chunks, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, i, row_chunk[0], row_chunk[1], num_rows_per_write, verbose) for i, row_chunk in enumerate(row_chunk_indices))
 
         line_lengths_dict = {}
         for x in line_lengths_dicts:
@@ -284,7 +401,6 @@ def get_line_lengths_dict(delimited_file_path, tmp_dir_path_chunks, delimiter, c
     return line_lengths_dict
 
 def write_rows_chunk(delimited_file_path, tmp_dir_path, delimiter, comment_prefix, compression_type, column_sizes, compression_dicts, chunk_number, start_index, end_index, num_rows_per_write, verbose):
-    #max_line_size = 0
     line_lengths_dict = {}
 
     if compression_type == "zstd":
@@ -326,23 +442,9 @@ def write_rows_chunk(delimited_file_path, tmp_dir_path, delimiter, comment_prefi
                 out_line = b"".join(out_items)
 
                 if compression_type == "zstd":
-                    #TODO
-                    # testing = False
-                    # if out_line.startswith(b"Row2    B") or out_line.startswith(b"Row1    B") or out_line.startswith(b"Row3    "):
-                    #     # print("build test before")
-                    #     # print(out_line)
-                    #     testing = True
                     out_line = compressor.compress(out_line)
-                    # if testing:
-                    #     print("build test after")
-                    #     print(out_line)
-                    #     decompressor = ZstdDecompressor()
-                    #     print(decompressor.decompress(out_line))
 
-                #line_size = len(out_line)
                 line_lengths_dict[line_index] = len(out_line)
-                #max_line_size = max([max_line_size, line_size])
-
                 out_lines.append(out_line)
 
                 if len(out_lines) % num_rows_per_write == 0:
@@ -353,7 +455,6 @@ def write_rows_chunk(delimited_file_path, tmp_dir_path, delimiter, comment_prefi
             if len(out_lines) > 0:
                 chunk_file.write(b"".join(out_lines))
 
-    #return max_line_size
     return line_lengths_dict
 
 def exclude_comments_and_header(in_file, comment_prefix):
@@ -454,28 +555,28 @@ def build_one_column_index(f4_file_path, index_column, tmp_dir_path, verbose, cu
     index_column_encoded = index_column.encode()
 
     print_message(f"Getting column meta information for {index_column} index for {f4_file_path}.", verbose)
-    file_data = initialize(f4_file_path)
-    select_columns, column_type_dict, column_coords_dict, bigram_size_dict = get_column_meta(file_data, set([index_column_encoded]), [])
+    with initialize(f4_file_path) as file_data:
+        select_columns, column_type_dict, column_coords_dict, bigram_size_dict = get_column_meta(file_data, set([index_column_encoded]), [])
 
-    index_column_type = column_type_dict[index_column_encoded]
-    coords = column_coords_dict[index_column_encoded]
-    values_positions = []
+        index_column_type = column_type_dict[index_column_encoded]
+        coords = column_coords_dict[index_column_encoded]
+        values_positions = []
 
-    parse_function = get_parse_row_value_function(file_data)
+        parse_function = get_parse_row_value_function(file_data)
 
-    print_message(f"Parsing values and positions for {index_column} index for {f4_file_path}.", verbose)
-    for row_index in range(file_data.stat_dict["num_rows"]):
-        value = parse_function(file_data, row_index, coords, bigram_size_dict=bigram_size_dict, column_name=index_column_encoded)
-        values_positions.append([value, row_index])
+        print_message(f"Parsing values and positions for {index_column} index for {f4_file_path}.", verbose)
+        for row_index in range(file_data.stat_dict["num_rows"]):
+            value = parse_function(file_data, row_index, coords, bigram_size_dict=bigram_size_dict, column_name=index_column_encoded)
+            values_positions.append([value, row_index])
 
-    print_message(f"Building index file for {index_column} index for {f4_file_path}.", verbose)
-    customize_index_values_positions(values_positions, [index_column_type], sort_first_column, custom_index_function)
-    write_index_files(values_positions, tmp_dir_path_data, tmp_dir_path_other)
+        print_message(f"Building index file for {index_column} index for {f4_file_path}.", verbose)
+        customize_index_values_positions(values_positions, [index_column_type], sort_first_column, custom_index_function)
+        write_index_files(values_positions, tmp_dir_path_data, tmp_dir_path_other)
 
-    index_file_path = get_index_file_path(f4_file_path, index_column, custom_index_function)
-    combine_into_single_file(tmp_dir_path_data, tmp_dir_path_other, index_file_path)
+        index_file_path = get_index_file_path(f4_file_path, index_column, custom_index_function)
+        combine_into_single_file(tmp_dir_path_data, tmp_dir_path_other, index_file_path)
 
-    print_message(f"Done building index file for {index_column} index for {f4_file_path}.", verbose)
+        print_message(f"Done building index file for {index_column} index for {f4_file_path}.", verbose)
 
 # TODO: Combine this function with the above one and make it generic enough to handle indexes with more columns.
 def build_two_column_index(f4_file_path, index_column_1, index_column_2, tmp_dir_path, verbose):
@@ -495,35 +596,33 @@ def build_two_column_index(f4_file_path, index_column_1, index_column_2, tmp_dir
     index_column_2_encoded = index_column_2.encode()
 
     print_message(f"Getting column meta information for {index_name} index.", verbose)
-    file_data = initialize(f4_file_path)
-    select_columns, column_type_dict, column_coords_dict, bigram_size_dict = get_column_meta(file_data, set([index_column_1_encoded, index_column_2_encoded]), [])
-    # TODO: Add logic to verify that index_column_1 and index_column_2 are valid.
+    with initialize(f4_file_path) as file_data:
+        select_columns, column_type_dict, column_coords_dict, bigram_size_dict = get_column_meta(file_data, set([index_column_1_encoded, index_column_2_encoded]), [])
+        # TODO: Add logic to verify that index_column_1 and index_column_2 are valid.
 
-    #line_length = file_data.stat_dict["ll"]
-    index_column_1_type = column_type_dict[index_column_1_encoded]
-    index_column_2_type = column_type_dict[index_column_2_encoded]
-    coords_1 = column_coords_dict[index_column_1_encoded]
-    coords_2 = column_coords_dict[index_column_2_encoded]
+        index_column_1_type = column_type_dict[index_column_1_encoded]
+        index_column_2_type = column_type_dict[index_column_2_encoded]
+        coords_1 = column_coords_dict[index_column_1_encoded]
+        coords_2 = column_coords_dict[index_column_2_encoded]
 
-    #decompressor = get_decompressor(decompression_type, decompressor)
-    parse_function = get_parse_row_value_function(file_data)
+        parse_function = get_parse_row_value_function(file_data)
 
-    values_positions = []
-    print_message(f"Parsing values and positions for {index_name} index.", verbose)
-    for row_index in range(file_data.stat_dict["num_rows"]):
-        value_1 = parse_function(file_data, row_index, coords_1, bigram_size_dict=bigram_size_dict, column_name=index_column_1_encoded)
-        value_2 = parse_function(file_data, row_index, coords_2, bigram_size_dict=bigram_size_dict, column_name=index_column_2_encoded)
+        values_positions = []
+        print_message(f"Parsing values and positions for {index_name} index.", verbose)
+        for row_index in range(file_data.stat_dict["num_rows"]):
+            value_1 = parse_function(file_data, row_index, coords_1, bigram_size_dict=bigram_size_dict, column_name=index_column_1_encoded)
+            value_2 = parse_function(file_data, row_index, coords_2, bigram_size_dict=bigram_size_dict, column_name=index_column_2_encoded)
 
-        values_positions.append([value_1, value_2, row_index])
+            values_positions.append([value_1, value_2, row_index])
 
-    print_message(f"Building index file for {index_name}.", verbose)
-    customize_index_values_positions(values_positions, [index_column_1_type, index_column_2_type], sort_first_two_columns, do_nothing)
-    write_index_files(values_positions, tmp_dir_path_data, tmp_dir_path_other)
+        print_message(f"Building index file for {index_name}.", verbose)
+        customize_index_values_positions(values_positions, [index_column_1_type, index_column_2_type], sort_first_two_columns, do_nothing)
+        write_index_files(values_positions, tmp_dir_path_data, tmp_dir_path_other)
 
-    index_file_path = get_index_file_path(f4_file_path, index_name)
-    combine_into_single_file(tmp_dir_path_data, tmp_dir_path_other, index_file_path)
+        index_file_path = get_index_file_path(f4_file_path, index_name)
+        combine_into_single_file(tmp_dir_path_data, tmp_dir_path_other, index_file_path)
 
-    print_message(f"Done building two-column index file for {index_name}.", verbose)
+        print_message(f"Done building two-column index file for {index_name}.", verbose)
 
 def customize_index_values_positions(values_positions, column_types, sort_function, custom_index_function):
     # Iterate through each "column" except the last one (which has row_indices) and convert the data.
@@ -591,4 +690,19 @@ def prepare_tmp_dirs(tmp_dir_path):
     makedirs(tmp_dir_path_outputs, exist_ok=True)
     makedirs(tmp_dir_path_indexes, exist_ok=True)
 
-    return tmp_dir_path_chunks, tmp_dir_path_outputs, tmp_dir_path_indexes
\ No newline at end of file
+    return tmp_dir_path_chunks, tmp_dir_path_outputs, tmp_dir_path_indexes
+
+def save_transposed_line_to_temp(data_file_path, tmp_file_path, col_names, col_coords, bigram_size_dict, tmp_dir_path, verbose):
+    print_message("got here6")
+    with initialize(data_file_path) as file_data:
+        print_message(f"Saving transposed lines to temp file at {tmp_file_path}", verbose)
+
+        #TODO: It would be better to save in zstandard format. But the Python package
+        #      doesn't support reading line by line. Come up with a better way than gzip,
+        #      which is very slow.
+        with gzip.open(tmp_file_path, "wb", compresslevel=1) as tmp_file:
+            for col_index in range(len(col_names)):
+                col_name = col_names[col_index]
+
+                col_values = [col_name] + parse_values_in_column(file_data, col_name, col_coords[col_index], bigram_size_dict)
+                tmp_file.write(b"\t".join(col_values) + b"\n")
\ No newline at end of file
