@@ -1,12 +1,10 @@
-#TODO:
-#from concurrent.futures import ThreadPoolExecutor, as_completed
+from contextlib import contextmanager
 from copy import deepcopy
 from datetime import datetime
 from glob import glob
 import gzip
 from fastnumbers import isint, isfloat, fast_int, fast_float
 from itertools import chain
-from joblib import Parallel, delayed
 from math import ceil, log
 from mmap import mmap, PROT_READ
 from msgspec import msgpack
@@ -23,10 +21,6 @@ FILE_KEY_ABBREVIATIONS_STATS = {"mccl": 3, "mctl": 6, "cnmccl": 10, "cnll": 11}
 FILE_KEY_ABBREVIATIONS_OTHER = {"ll": 4, "cmpr": 7}
 FILE_KEY_ABBREVIATIONS_NOCACHE = {"data": 1, "cc": 2, "ct": 5, "cndata": 8, "cncc": 9}
 
-def open_read_file(file_path, file_extension=""):
-    the_file = open(file_path + file_extension, 'rb')
-    return mmap(the_file.fileno(), 0, prot=PROT_READ)
-
 def read_str_from_file(file_path, file_extension=""):
     with open(file_path + file_extension, 'rb') as the_file:
         return the_file.read()
@@ -63,7 +57,8 @@ def format_column_items(the_list, max_value_length, suffix=""):
 
 def print_message(message, verbose=False):
     if verbose:
-        print(f"{message} - {datetime.now().strftime('%d/%m/%Y %H:%M:%S.%f')}")
+        sys.stderr.buffer.write(f"{message} - {datetime.now().strftime('%d/%m/%Y %H:%M:%S.%f')}")
+        sys.stderr.flush()
 
 def decode_string(x):
     return x.decode()
@@ -93,6 +88,9 @@ def reverse_string(s):
 def get_delimited_file_handle(file_path):
     if file_path.endswith(".gz"):
         return gzip.open(file_path)
+    elif file_path.endswith(".zstd"):
+        with open(file_path, 'rb') as fh:
+            return ZstdCompressor(level=1).stream_reader(fh)
     else:
         return open(file_path, 'rb')
 
@@ -151,7 +149,7 @@ def remove_tmp_dir(tmp_dir_path):
             print(e)
             pass
 
-def combine_into_single_file(tmp_dir_path_chunks, tmp_dir_path_outputs, f4_file_path, num_threads=1, num_rows_per_write=1):
+def combine_into_single_file(tmp_dir_path_chunks, tmp_dir_path_outputs, f4_file_path, num_parallel=1, num_rows_per_write=1):
     def _create_file_map(start_end_positions):
         start_end_dict = {}
 
@@ -203,15 +201,15 @@ def combine_into_single_file(tmp_dir_path_chunks, tmp_dir_path_outputs, f4_file_
     with open(f4_file_path, "wb") as f4_file:
         f4_file.write(file_map_serialized)
 
-        _add_data_chunks(tmp_dir_path_chunks, f4_file, num_threads, num_rows_per_write)
+        _add_data_chunks(tmp_dir_path_chunks, f4_file, num_parallel, num_rows_per_write)
 
         for other_file_path in other_files:
             f4_file.write(read_str_from_file(other_file_path))
 
-def _add_data_chunks(tmp_dir_path_chunks, out_file_handle, num_threads, num_rows_per_write):
+def _add_data_chunks(tmp_dir_path_chunks, out_file_handle, num_parallel, num_rows_per_write):
     out_lines = []
 
-    for i in range(num_threads):
+    for i in range(num_parallel):
         chunk_file_path = f"{tmp_dir_path_chunks}{i}"
 
         if not path.exists(chunk_file_path):
@@ -234,4 +232,19 @@ def get_index_file_path(data_file_path, index_name, custom_index_function=do_not
     if custom_index_function != do_nothing:
         index_file_path_extension = f"{index_file_path_extension}_{custom_index_function.__name__}"
 
-    return f"{data_file_path}{index_file_path_extension}"
\ No newline at end of file
+    return f"{data_file_path}{index_file_path_extension}"
+
+def split_integer_list_into_chunks(int_list, num_parallel):
+    items_per_chunk = ceil(len(int_list) / num_parallel)
+
+    return_indices = list()
+
+    for an_int in int_list:
+        return_indices.append(an_int)
+
+        if len(return_indices) == items_per_chunk:
+            yield return_indices
+            return_indices = list()
+
+    if len(return_indices) > 0:
+        yield return_indices
\ No newline at end of file
